% =====================================================================================
%
%  This program is free software: you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation, either version 3 of the License, or
%  (at your option) any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% -------------------------------------------------------------------------------------
%  The primary purpose in distributing this source code is to enable 
%  readers to reproduce the numerical results reported in our manuscript
%  on optimal control under constraints on reset-renewable resources.
%  See  http://arxiv.org/abs/1110.6221
%
%  Please see Readme.txt for instructions on configuring/running this program.
% -------------------------------------------------------------------------------------
%  Problem statement by R. Takei.
%  The idea of the algorithm by A. Vladimirsky.
%  The numerical method development by R. Takei & A. Vladimirsky.
%  Implementation by R. Takei.
%  
%  The method implemented in these files is not as general as the description 
%  in the paper and relies on several implicit assumptions:
%
%		1) the dynamics is isotropic and the primary cost is identically = 1 on Omega;
%		2) the secondary cost on U is identically = 1;
%		3) the discretization parameters (h and deltaB) satisfy
%			a CFL-type condition:  deltaB * f(x) \leq h.
%
%  These assumptions are not essential for the approach, but simplify the implementation.
%
%  To handle inhomogenious/anisotropic secondary costs and arbitrary deltaB/h ratios,
%  an alternative implementation was developed & tested by W. Chen, Z. Clawson, 
%  and S. Kirov based on the code for the U=Omega case previously developed by A. Kumar.
%  This more general implementation is not included in the current distribution.
% =====================================================================================
% This work was supported by
% ONR grants N0014-03-1-0071, N00014-07-1-0810, N00014-08-1-1119,
% DOE grant DE-FG02-05ER25710,
% ARO MURI through Rice University and the CHASE MURI grant 556016,
% National Science Foundation grants DMS-0514487 and DMS-1016150,
% and the NSF sponsored REU'2010 at Cornell University.
% =====================================================================================

clear all; close all;

% compile mex code
% mex solveEikonal.cpp; 
% mex solveConstrainedEikonal.cpp; 
% mex solveOtherOptEikonal.cpp; 
% mex computeVisibility.cpp;

% constraint on budget
B = 0.4;

% parameters
N  = 200;
L  = 1;           % [-L,L]^2 physical space
h  = 2*L / (N-1); % grid spacing  
hb = 0.8*h;       % grid spacing in b w.r.t. h
TOL = 1e-8;       % tolerance of outer iteration

% set to 1 to show intermediate plots, 0 for final plots only
P_INT = 1;

% CHOOSE TEST CASE NUMBER TO LOAD PARAMETERS:
test = 3;
% test = 1: generic example
% test = 2: high cost in safe set
% test = 3: 2-d sinusoidal cost function
% test = 4: square chain of square
% test = 5: for convergence test
% test = 6: visibility example
[R,OB,f,xT,yT,x,y] = testcases(test,L,N);
%... or you can manually override the values yourself.
%f = ...

% SET THE plotOptimalPath = 1 TO SHOW OPTIMAL PATH, 0 OTHERWISE.
plotOptimalPath = 1; 
% ENTER INITIAL POSITION HERE  
%(but make sure there exists a feasible path from there to the target)
X0 = 0.8;
Y0 = 0.5;
% for test 6, use the following initial point    
if(test == 6)
    X0 = 0.7;
    Y0 = -0.8;
end


%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%% DO NOT MODIFY BELOW %%%%%%%%%%%%%%%%
% re-adjust to give integer number of mesh pts in the b direction
Nb = round(B/(hb) + 1); 
hb = B / (Nb-1);

% target
TargetX = round((N-1)/(2*L)*(xT+L)+1); % x target in pixels
TargetY = round((N-1)/(2*L)*(yT+L)+1); % y target in pixels

if(1)
    % plot
    figure(99),clf;
    pcolor(x,y,f); shading interp; colormap(autumn); colorbar; hold on;
    contour(x,y,R,[0 0],'b-','LineWidth',4); 
    %plot(TargetX,TargetY,'k+','LineWidth',4);
    plot(xT,yT,'k+','MarkerSize',10);
    title('Speed function');
    axis square; 

    figure(98),clf;
    contourf(x,y,-R,[0 0],'k-','LineWidth',4); colormap(gray); hold on;
    %contour(x,y,R,[0 0],'b-','LineWidth',4); 
    plot(xT,yT,'k+','MarkerSize',10);
    title('S (shaded) & U (unshaded)');
    axis square; 
end

if(test == 6)
    figure(98),clf;
    contourf(x,y,(R-OB),[5 0 -1 -5],'k-','LineWidth',2); colormap(gray); hold on;
    %contourf(x,y,OB,[0.5 0.5],'k-','LineWidth',4);
    plot(0.8,0.8,'k+','MarkerSize',10);
    plot(-0.5,0.9,'ks','MarkerSize',10);
    plot(X0,Y0,'ks','MarkerSize',10);
    title('S (shaded) & U (unshaded)');
    axis square;  drawnow;
end

%% STEP 0: Initializations 
% check if target is in bad region; if so, extend good region
if (R(TargetY,TargetX) > 0)
    R(TargetY,TargetX) = -1; % creat ab one pixel 'island' of good set
end

% define safe and unsafe regions
SAFE   = double( R <= 0); % SAFE region
UNSAFE = double( R >  0); % UNSAFE region

% initialize functions
W1 = 1000*ones(N,N,Nb); % value function on Omega_UNSAFE
W2 = 1000*ones(N,N); % value function on SAFE

% set boundary conditions (zero b.c. at target point)
W1(TargetY, TargetX, :) = 0;
W2(TargetY, TargetX)    = 0;

%% MAIN LOOP
% store old values for convergence criterion
W1_old = W1; 
W2_old = W2;

tic; % count time

% unconstrained value function
U = 1000*ones(N,N); % initial condition for MFL
bdryCond = -1*ones(N,N);
bdryCond(TargetY, TargetX) = 0;
obs = OB;
numIter = 10;
solveEikonal(U, bdryCond, obs, numIter, h, f);

for m = 1:20 % this is the maximum number of iterations

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
    %% PHASE I
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    %% STEP 1: find set in the reachable set in the unsafe region
    bdryCond = (W2 < 500).*SAFE - 1 + 1000*(W2 >= 500).*SAFE; %(W2 < 500)-1; % find reachable set in good set
    obs = OB;        % can add obstacles later
    numIter = 10;            % number of sweeping interations
    
    V = 1000*ones(N,N); % initial condition for MFL
    solveEikonal(V, bdryCond, obs, numIter, h, f);
    
    
    %% STEP 3: compute "otherwise optimal" value function
    udist = V;                   % other optimal value function
    ubdry = W2 .* SAFE - UNSAFE; % boundary condition
    obs = max(double(V > B),OB);         % areas that cannot be reached 
    numIter = 20;                % number of sweeping interations
    tau = h;                     % step size

    Ut = 1000*ones(N,N); % initial condition
    solveOtherOptEikonal(Ut, udist, ubdry, obs, numIter, tau, h, f);
    
    %% STEP 4: set boundary condition (U tilde) on MFL
    % ... and infinity below! (changed Aug 5, 2001 by rrtakei)
    bdryMFL = -1*ones(N,N,Nb); % initialize MFL boundary values
    
    for k = 1:Nb
        % points in space where MFL = b, approximately
        on_MFL =    ((floor(V/hb)+1) == k).*(V <= B) > 0;
        
        % points in space where MFL > b
        below_MFL = ((floor(V/hb)+1)  > k) > 0;
        
        % initialize auxiliary function
        bdryMFL_slice = -1*ones(N,N);
        % set Ut values on MFL and in the safe set
        bdryMFL_slice(on_MFL) = Ut(on_MFL);
        bdryMFL_slice(SAFE > 0) = Ut(SAFE > 0);
        % set inifinity below MFL
        bdryMFL_slice(below_MFL) = 1000;
        
        % plug into k-th slice
        bdryMFL(:,:,k) = bdryMFL_slice;        
        
        % test
%         figure(50),clf;
%         pcolor(x,y,bdryMFL_slice.*(bdryMFL_slice < 500).*(bdryMFL_slice>-1));
%         title(k); axis square; drawnow;
    end

    %% STEP 5: solve the constrained eikonal PDE
    URbdry = -1*ones(N,N,Nb);
    for k = 1:Nb
        URbdry(:,:,k) = double(W2 .* SAFE - UNSAFE);
    end
    obs = max(double(V > B),OB); 
    numIter = 1;            % number of sweeping interations
    
    W1 = 1000*ones(N,N,Nb);
    solveConstrainedEikonal(W1, URbdry, bdryMFL, obs, numIter, h, B, f, U);
    
    %% STEP 6: extract minimum time value function for constrained problem
    W1_top = W1(:,:,Nb);
    maxW1 = max(max(W1_top .* (V < B*0.97)));
    W_top = (W1_top .* UNSAFE) + (W2 .* SAFE);
    
    if(P_INT)
        figure(1),clf;
        if (test == 4)
            contourf(x,y, log(W_top+.01),log([linspace(0,0.4,40),linspace(0.4,3,2)]+.01)); 
            colormap('gray'); hold on;
        else
            contourf(x,y, W_top, linspace(0,maxW1,60)); colormap('summer'); hold on;
        end
        contour(x,y, R,  [0 0],'b-','LineWidth',4); V_plot = V; V_plot(SAFE>0) = NaN; 
        contour(x,y, V_plot, [B B],'k-','LineWidth',4);
        axis square; title(['iteration ',num2str(m),', Phase I']); drawnow;
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Phase II
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    %% STEP 7:
    bdryCond = (W1_top .* UNSAFE) - (SAFE); % boundary condition
    bdryCond(TargetY, TargetX) = 0;         % boundary condition at target
    obs = OB;                       % obstacle, can add later
    numIter = 10;                           % number of sweeping interations  

    W2 = W1_top .* SAFE + 1000 * UNSAFE;        % initialize value func with known values
    solveEikonal(W2, bdryCond, obs, numIter, h, f);
    
    % for plotting purposes only
    W_top_old = W_top; % for testing only
    W_top = (W1_top .* UNSAFE) + (W2 .* SAFE);
%     figure(55),clf;
%     surf(x,y,abs(W_top-W_top_old));
%     title(['iteration ',num2str(m)]);
%     drawnow;
    maxW = max(max(W_top .* (V < B*0.97)));
    V_plot = V; V_plot(SAFE>0) = NaN; 
    if(P_INT)
        figure(2),clf;
        if (test == 4)
            contourf(x,y, log(W_top+.01),log([linspace(0,0.4,40),linspace(0.4,3,2)]+.01)); 
    colormap('gray'); hold on;
        else
            contourf(x,y, W_top, linspace(0,maxW,60)); colormap('summer'); hold on;
        end
        
        contour(x,y, R,  [0 0],'b-','LineWidth',4);
        contour(x,y, V_plot, [B B],'k-','LineWidth',4);
        axis square; title(['iteration ',num2str(m),', Phase II']); drawnow;
    end
    
    %% STEP 8: break main loop if convergence is met
    W1_error = max(max(max(abs(W1 - W1_old),[],3) .* UNSAFE));
    W2_error = max(max(    abs(W2 - W2_old      ) .*   SAFE));
    if ( max(W1_error, W2_error) < TOL)
        disp(['Iteration ',num2str(m),'. W1 Error = ',num2str(W1_error),', W2 Error = ',num2str(W2_error) ]);
        disp(['Convergence met (TOL = ',num2str(TOL),').']);
        break;
    else
        disp(['Iteration ',num2str(m),'. W1 Error = ',num2str(W1_error),', W2 Error = ',num2str(W2_error) ]);
    end
    
    W1_old = W1;
    W2_old = W2;

    %pause;
end % END MAIN LOOP
toc;

figure(1),clf;
if (test == 4)
    contourf(x,y, log(W_top+.01),log([linspace(0,0.4,40),linspace(0.4,3,2)]+.01)); 
    colormap('gray'); hold on;
else
    contourf(x,y, W_top, linspace(0,maxW1,60)); colormap('summer'); hold on;
end
contour(x,y, R,  [0 0],'b-','LineWidth',4); V_plot = V; V_plot(SAFE>0) = NaN; 
contour(x,y, V_plot, [B B],'k-','LineWidth',2);
contour(x,y, OB, .5*[1 1],'k-','LineWidth',4); 
axis square; title(['iteration ',num2str(m),', Phase I']); drawnow;


%% STEP 9: compute optimal path
if(plotOptimalPath)    
    % compute and plot optimal path
    figure(3),clf;
    contourf(x,y, W2, linspace(0,maxW,60)); colormap('summer'); hold on;
    contour(x,y, R,  [0 0],'b-','LineWidth',4);
    contour(x,y, OB, .5*[1 1],'k-','LineWidth',4); 
    contour(x,y, V_plot, [B B],'k-','LineWidth',4);
    axis square; drawnow; %title('Optimal path'); drawnow;
    % call ComputePath.m
    fig_num = 3;
    u4 = W1; ComputePath;
end