
% THIS CODE IS TO BE EXECUTED WITHIN 'ContaminatedEikonal.m'!

% =====================================================================================
%
%  This program is free software: you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation, either version 3 of the License, or
%  (at your option) any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% -------------------------------------------------------------------------------------
%  The primary purpose in distributing this source code is to enable 
%  readers to reproduce the numerical results reported in our manuscript
%  on optimal control under constraints on reset-renewable resources.
%  See  http://arxiv.org/abs/1110.6221
%
%  Please see Readme.txt for instructions on configuring/running this program.
% -------------------------------------------------------------------------------------
%  Problem statement by R. Takei.
%  The idea of the algorithm by A. Vladimirsky.
%  The numerical method development by R. Takei & A. Vladimirsky.
%  Implementation by R. Takei.
%  
%  The method implemented in these files is not as general as the description 
%  in the paper and relies on several implicit assumptions:
%
%		1) the dynamics is isotropic and the primary cost is identically = 1 on Omega;
%		2) the secondary cost on U is identically = 1;
%		3) the discretization parameters (h and deltaB) satisfy
%			a CFL-type condition:  deltaB * f(x) \leq h.
%
%  These assumptions are not essential for the approach, but simplify the implementation.
%
%  To handle inhomogenious/anisotropic secondary costs and arbitrary deltaB/h ratios,
%  an alternative implementation was developed & tested by W. Chen, Z. Clawson, 
%  and S. Kirov based on the code for the U=Omega case previously developed by A. Kumar.
%  This more general implementation is not included in the current distribution.
% =====================================================================================
% This work was supported by
% ONR grants N0014-03-1-0071, N00014-07-1-0810, N00014-08-1-1119,
% DOE grant DE-FG02-05ER25710,
% ARO MURI through Rice University and the CHASE MURI grant 556016,
% National Science Foundation grants DMS-0514487 and DMS-1016150,
% and the NSF sponsored REU'2010 at Cornell University.
% =====================================================================================


% THIS CODE IS TO BE EXECUTED WITHIN 'ContaminatedEikonal.m'!

% initial point in pixels
x0 = round((N-1)/(2*L)*(X0+L)+1); 
y0 = round((N-1)/(2*L)*(Y0+L)+1);

% initial budget = B
b0 = B;

% arc length 
arcLength = 0.5;

% tolerance
TOL = arcLength*3;

% start path vector
xv = x0;
yv = y0;
bv = b0;

% loop while goal is not within a tolerance
while (sqrt( (xv(end)-TargetX)^2 + (yv(end)-TargetY)^2 ) > TOL)
    
    % STEP 1: compute gradient by nearest neighbor approximation
    xn = round(xv(end));
    yn = round(yv(end));
    bnn = round(bv(end)*(Nb-1)/B + 1);
    
    if(bnn==0)
        bnn = 1;
    end
    
    xnp = xn + 1;
    xnn = xn - 1;
    ynp = yn + 1;
    ynn = yn - 1;
    
    % check if its too close to the edge
    if (xn < 2)||(xn > N-1)||(yn < 2)||(yn > N-1)
        disp('WARNING: Path outside domain or too close to edge'); 
        break;
    end
    
    % compute gradient
    gradx = u4(yn , xnp, bnn) - u4(yn , xnn, bnn); 
    grady = u4(ynp, xn , bnn) - u4(ynn, xn , bnn); 
    
    grad_norm = sqrt(gradx^2 + grady^2);
    
    if (grad_norm == 0)
        disp('ERROR: Initial point is not reachable.'); 
        xv = -100;
        yv = -100;
        break;
    end
    
    % STEP 2: advance point in space (x,y)
    xv = [xv, xv(end)-gradx/grad_norm*arcLength];  %#ok<AGROW>
    yv = [yv, yv(end)-grady/grad_norm*arcLength];  %#ok<AGROW>
    
    % STEP 3: check region and advance budget b
    xR = (xv(end) - 1)*2/(N-1) - 1;
    yR = (yv(end) - 1)*2/(N-1) - 1;
    
    if (R(round(yv(end)),round(xv(end))) <= 0)
        % if in the good region, push up to max budget B
        bv = [bv, B]; %#ok<AGROW>
    else 
        % if in bad region, decrease budget
        bv = [bv, bv(end)-h*arcLength]; %#ok<AGROW>
    end
end

%scale back down to [-L,L] space:
xvs = (xv-1)*2*L/(N-1) - 1;
yvs = (yv-1)*2*L/(N-1) - 1;

figure(fig_num),hold on;
plot(xvs,yvs,'r--','LineWidth',2);
plot(X0,Y0,'bo','MarkerSize',10);

%Uncomment below to plot without contour of value function
% figure(99),clf;
% contour(x,y,R,[0 0],'b-','LineWidth',4); hold on;
% plot(TargetX,TargetY,'k+','LineWidth',4);
% plot(X0,X0,'mo','LineWidth',4);
% contour(x,y, u2, [B B],'k-','LineWidth',4);
% plot(xvs,yvs,'r.-');
% axis square;
% legend('region bdry','goal','start','feasible bdry','optimal path','Location','BestOutside');

% % plot 3d plot including budget value
uu = ones(N,N,Nb);
uu(:,:,1) = R;
[xx,yy,bb] = meshgrid(linspace(-L,L,N),linspace(-L,L,N),linspace(0,B,Nb));
% 
figure(97),clf;
plot3(xvs,yvs,bv,'r.-'); axis([-L L -L L 0 B]); axis square; hold on;
isosurface(xx,yy,bb,uu,0,'b'); colormap(jet);
drawnow;
