   /*
  * =====================================================================================
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation, either version 3 of the License, or
  *  (at your option) any later version.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *
  * -------------------------------------------------------------------------------------
  *  The primary purpose in distributing this source code is to enable 
  *  readers to reproduce the numerical results reported in our manuscript
  *  on optimal control under constraints on reset-renewable resources.
  *  See  http://arxiv.org/abs/1110.6221
  *
  *  Please see Readme.txt for instructions on configuring/running this program.
  * -------------------------------------------------------------------------------------
  *  Problem statement by R. Takei.
  *  The idea of the algorithm by A. Vladimirsky.
  *  The numerical method development by R. Takei & A. Vladimirsky.
  *  Implementation by R. Takei.
  *  
  *  The method implemented in these files is not as general as the description 
  *  in the paper and relies on several implicit assumptions:
  *
  *		1) the dynamics is isotropic and the primary cost is identically = 1 on Omega;
  *		2) the secondary cost on U is identically = 1;
  *		3) the discretization parameters (h and deltaB) satisfy
  *			a CFL-type condition:  deltaB * f(x) \leq h.
  *
  *  These assumptions are not essential for the approach, but simplify the implementation.
  *
  *  To handle inhomogenious/anisotropic secondary costs and arbitrary deltaB/h ratios,
  *  an alternative implementation was developed & tested by W. Chen, Z. Clawson, 
  *  and S. Kirov based on the code for the U=Omega case previously developed by A. Kumar.
  *  This more general implementation is not included in the current distribution.
  * =====================================================================================
  */

#define min(a,b) (((a) < (b)) ? (a) : (b))
#define max(a,b) (((a) > (b)) ? (a) : (b))
#define signFunc(a) (((a) < 0) ? -1.0 : 1.0)
#define abs(a) (((a) < 0) ? (-a) : (a))
#define PI 3.14159265358979323846
//#define SQRT2 1.414213562373
//#define SQRT5 2.236067977499

void signDistPDE(double** u,int N,double L,int NumIter);
double updateOneStep(int i,int j,double** u,double h, int N);
double sweepSignDist(double** u,double** obs,int N);
double SweepOneStep(int i,int j,double** u,double** obs,int N);
//double updateOneStepSSO(int i, int j, double** u, double h, int N);
double Hh(double u, double h);
double sigmah(double u, double h);
double deltah(double u, double h);
double gradh(int i, int j, double **u, double h, int N) ;
double fh(int i, int j, double **u, double h, int N);
double Lh(int i, int j, double **u, double h, int N);
double dphi_p(double a, double b, double c, double d, double h);
double dphi_m(double a, double b, double c, double d, double h); 
double r(double a, double b, double c, double d, double h);
double SweepOneStep_HO(int i,int j,double** u,double** obs,int N);
double FSM_HO(double** u,double** obs,int N);

void signDistPDE(double** u,int N,double L,int NumIter) {
    double **v,**f;
    int n,i,j;
    double h = 2*L/(N-1);
    
    // allocate memory for auxiliary matrix v, initial f
    v = (double **) malloc ( N * sizeof(double*));
	for (i=0;i<N;i++) {
        v[i] = (double *) malloc (N * sizeof(double));
    }

    
    // Iterate 
    for (n=0; n<NumIter; n++) 
    {
        // traverse along each inner grid point 
        for (i=1; i<N-1; i++) 
            for (j=1; j<N-1; j++)
                v[i][j] = updateOneStep(i,j,u,h,N);
        
        
        // Neumann boundary conditions
        for (i=0; i<N; i++) {
            v[i][0]   = v[i][1]   +h;
            v[i][N-1] = v[i][N-2] +h;
        }
        for (j=0; j<N; j++) {
            v[0][j]   = v[1][j]   +h;
            v[N-1][j] = v[N-2][j] +h;
        }
        
        
        // update u
        for (i=0; i<N; i++) 
            for (j=0; j<N; j++)
                u[i][j] = v[i][j];
    }
    
    // delete v;
	for(i=0; i < N; i++) 
		free(v[i]);
	free(v);
    
        
}

/*
double updateOneStepSSO(int i, int j, double** u, double h, int N) {
    double dt = h*0.5;     // time step 
    double u0 = u[i][j];
    double k1,k2;
    
    k1 = dt*Lh(i, j, u,        h, N);
    k2 = dt*Lh(i, j, u+0.5*k1, h, N);
    
    // RK-2
    return u0 + k2;
}
*/

double Hh(double u, double h) {
    if (u > h) {
        return 1;
    } else if (u < -h) {
        return 0;
    } else {
        return 0.5*(1 + u/h + sin(PI*u/h)/PI);
    }
}

double sigmah(double u, double h) {
    return 2*(Hh(u,h) - 0.5);
}

double deltah(double u, double h) {
    return 0.5/h * (1 + cos(PI*u/h));
}


// first order
double gradh(int i, int j, double **u, double h, int N) {
    double u0, a,b,c,d;
    
    u0 = u[i][j];
    a = ( u0        - u[i][j-1] );
    b = ( u[i][j+1] - u0        );
    c = ( u0        - u[i-1][j] );
    d = ( u[i+1][j] - u0        );
    
    if (u0 >= 0) {
        a = max(a,0); b = min(b,0);
        c = max(c,0); d = min(d,0);
    } else {
        a = min(a,0); b = max(b,0);
        c = min(c,0); d = max(d,0); 
    }
    
    return sqrt(max(a*a,b*b) + max(c*c,d*d)) / h;
}


double fh(int i, int j, double **u, double h, int N) {
    double d = deltah(u[i][j], h);
    double g = gradh(i, j, u, h, N);
    
    return d*g;
}

double Lh(int i, int j, double **u, double h, int N) {
    double s = sigmah(u[i][j], h);
    double g = gradh(i, j, u, h, N);
    
    return s*(1-g);
}



double updateOneStep(int i, int j, double** u, double h, int N) {
    double S, G;
    double dt = h*0.5;     // time step 
    double u0 = u[i][j];
    
    //gradient
    G = gradh(i,j,u,h,N);
   
    // regularized speed of front
    S = sigmah(u0,h);
    
    return u0 - dt * S * (G - 1);
}



// engine
double sweepSignDist(double** u,double** obs,int N)
{
	int s1,s2,i,j;
    double utemp;
    double err = 0.0;
    
	for (s1=1; s1>=-1; s1-=2)
		for (s2=1; s2>=-1; s2-=2)
			for( i=(s1<0 ? N-1:0); (s1<0 ? i>=0:i<=N-1); i+=s1 )
				for( j=(s2<0 ? N-1:0); (s2<0 ? j>=0:j<=N-1); j+=s2 ) {
                    utemp = u[i][j];
                    if (utemp >= 0) {
                        u[i][j] = SweepOneStep(i,j,u,obs,N);
                        utemp = utemp - u[i][j];
                        utemp = abs(utemp);
                        err = max(err, utemp); //store max error
                    }
                }
                
    return err;
}

// helper function
double SweepOneStep(int i,int j,double** u,double** obs,int N)
{
    double u_ij, c_ij, uxp,uxm,uyp,uym, a,b,d, signU;
    double h = 2.0/(N-1); // for [-1,1]^2 domain!
    double u0 = u[i][j];
    
    if (obs[i][j] < 0.0) // in obstacle
    {
        // return 100
        u_ij = 100;
    }
    else if ((u0 < h) && (u0 > -h)) // if on narrow band boundary condition
    {
        // do nothing and return as is
        u_ij = u0;
    } 
    else if (u0 > 0 )
    {   
        // check top/bottom point on stencil
        if      (i==0)              { uym = 100; }
        else if (obs[i-1][j] < 0.0) { uym = 100; }
        else                        { uym = u[i-1][j]; }
        if      (i==N-1)            { uyp = 100 ; }
        else if (obs[i+1][j] < 0.0) { uyp = 100; }
        else                        { uyp = u[i+1][j]; }
        
        // check left/right point on stencil
        if      (j==0)              { uxm = 100; }
        else if (obs[i][j-1] < 0.0) { uxm = 100; }
        else                        { uxm = u[i][j-1]; }
        if      (j==N-1)            { uxp = 100; }
        else if (obs[i][j+1] < 0.0) { uxp = 100; }
        else                        { uxp = u[i][j+1]; }
        
        // slowness function
        c_ij = 1;
        
        // standard formula
        a = min(uxp,uxm);
        b = min(uyp,uym);
        d = 2*h*h*c_ij*c_ij-(a-b)*(a-b);
        if (d < 0) {
            u_ij = h*c_ij + min(a,b);
        } else {
            u_ij = 0.5*(a + b + sqrt(d)); 
            
            if (b > u_ij && a < u_ij) { 
                u_ij = h*c_ij + a; 
            } else if (a > u_ij && b < u_ij) { 
                u_ij = h*c_ij + b;
            } 
        }
        
        // check min with current u[i][j]
        u_ij = min(u_ij, u[i][j]);
    } else {
        // negative area, just return 0
        u_ij = 0;
    }
    
    return u_ij;
}


double FSM_HO(double** u,double** obs,int N) {
    int s1,s2,i,j;
    double utemp;
    double err = 0.0;
    
	for (s1=1; s1>=-1; s1-=2)
		for (s2=1; s2>=-1; s2-=2)
			for( i=(s1<0 ? N-1:0); (s1<0 ? i>=0:i<=N-1); i+=s1 )
				for( j=(s2<0 ? N-1:0); (s2<0 ? j>=0:j<=N-1); j+=s2 ) {
                    utemp = u[i][j];
                    if (utemp >= 0) {
                        u[i][j] = SweepOneStep_HO(i,j,u,obs,N);
                        utemp = utemp - u[i][j];
                        utemp = abs(utemp);
                        err = max(err, utemp); //store max error
                    }
                }
                
    return err;
}

double SweepOneStep_HO(int i,int j,double** u,double** obs,int N) {
    double c_ij, ux1,ux2, uy1,uy2, uXM,uYM, diff;
    double h = 2.0/(N-1); // for [-1,1]^2 domain!
    double u0 = u[i][j];
    
    if (obs[i][j] < 0.0) { // in obstacle
        // return 100
        return 100;
    } else if ((u0 < h) && (u0 > -h)) { // if on narrow band boundary condition
        // do nothing and return as is
        return u0;
    } else if ( (i<=1)||(i>=N-2)||(j<=1)||(j>=N-2) ) {
        return SweepOneStep(i,j,u,obs,N);
    } else if (u0 > 0) {   
        // check top/bottom point on stencil
        if ( (obs[i][j+1]<0)||(obs[i][j+2]<0)||(obs[i][j-1]<0)||(obs[i][j-2]<0) ){
            return SweepOneStep(i,j,u,obs,N);
        }
        if ( (obs[i+1][j]<0)||(obs[i+2][j]<0)||(obs[i-1][j]<0)||(obs[i-2][j]<0) ){
            return SweepOneStep(i,j,u,obs,N);
        }
        
        ux1 = u0 + h*dphi_p(u[i][j-1],u[i][j],  u[i][j+1],u[i][j+2],h);
        ux2 = u0 - h*dphi_m(u[i][j-2],u[i][j-1],u[i][j],  u[i][j+1],h);
         
        uy1 = u0 + h*dphi_p(u[i-1][j],u[i][j],  u[i+1][j],u[i+2][j],h);
        uy2 = u0 - h*dphi_m(u[i-2][j],u[i-1][j],u[i][j],  u[i+1][j],h);
        
        uXM = min(ux1, ux2);
        uYM = min(uy1, uy2);
        
        // slowness function
        c_ij = 1;
        
        diff = uXM - uYM;
        diff = abs(diff);
        if (diff >= c_ij*h) {
            return min(uXM, uYM) + c_ij*h;
        } else {
            return 0.5*(uXM+uYM+ sqrt(2*c_ij*c_ij*h*h - diff*diff));
        }
    } else {
        // negative area, just return 0
        return 0;
    }
}

double dphi_p(double a, double b, double c, double d, double h) {
    double rr, w, u;
    
    rr = r(a,b,c,d, h);
    w = 1 / (1 + 2*rr*rr);
    
    u = (1-w)*(c-a) - w*(d-4*c+3*b);
    return u / (2*h);
}

double dphi_m(double a, double b, double c, double d, double h) {
    double rr, w, u;
    
    rr = r(d,c,b,a, h);
    w = 1 / (1 + 2*rr*rr);
    
    u = (1-w)*(d-b) + w*(a-4*b+3*c);
    return u / (2*h);
}

double r(double a, double b, double c, double d, double h) {
    double z1,z2;
    
    z1 = h + (d - 2*c + b)*(d - 2*c + b);
    z2 = h + (c - 2*b + a)*(c - 2*b + a);
    
    return z1/z2;
}
