   /*
  * =====================================================================================
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation, either version 3 of the License, or
  *  (at your option) any later version.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *
  * -------------------------------------------------------------------------------------
  *  The primary purpose in distributing this source code is to enable 
  *  readers to reproduce the numerical results reported in our manuscript
  *  on optimal control under constraints on reset-renewable resources.
  *  See  http://arxiv.org/abs/1110.6221
  *
  *  Please see Readme.txt for instructions on configuring/running this program.
  * -------------------------------------------------------------------------------------
  *  Problem statement by R. Takei.
  *  The idea of the algorithm by A. Vladimirsky.
  *  The numerical method development by R. Takei & A. Vladimirsky.
  *  Implementation by R. Takei.
  *  
  *  The method implemented in these files is not as general as the description 
  *  in the paper and relies on several implicit assumptions:
  *
  *		1) the dynamics is isotropic and the primary cost is identically = 1 on Omega;
  *		2) the secondary cost on U is identically = 1;
  *		3) the discretization parameters (h and deltaB) satisfy
  *			a CFL-type condition:  deltaB * f(x) \leq h.
  *
  *  These assumptions are not essential for the approach, but simplify the implementation.
  *
  *  To handle inhomogenious/anisotropic secondary costs and arbitrary deltaB/h ratios,
  *  an alternative implementation was developed & tested by W. Chen, Z. Clawson, 
  *  and S. Kirov based on the code for the U=Omega case previously developed by A. Kumar.
  *  This more general implementation is not included in the current distribution.
  * =====================================================================================
  */

#define min(a,b) (((a) < (b)) ? (a) : (b))
#define max(a,b) (((a) > (b)) ? (a) : (b))

void singleVisFunc(double** visFuncTemp,int x00, int y00,double** obs,int N,double dx);
void singleShadowFunc(double** visFuncTemp,int x00, int y00,double** obs,int N,double dx);
void singleSweep(double** visFuncTemp,int x00,int y00,double** obs,int N,double dx,int sX,int sY);

void singleShadowFunc(double** visShadTemp,int x00, int y00,double** obs,int N,double dx) {
    int i,j;
    double psi_x, psi_y, v1, v2, h, r1,r2,rs;
    double **v;
    
    // allocate memory for auxiliary matrix v
    v = (double **) malloc ( N * sizeof(double*));
	for (i=0; i<N; i++)
        v[i] = (double *) malloc (N * sizeof(double));
    
    for(i=0; i<N; i++)
        for(j=0; j<N; j++) {
            v[i][j] = 1;
        }
    
    // auxiliary function for interior points...
    for(i=1; i<N-1; i++)
        for(j=1; j<N-1; j++) {
            psi_x = 0.5*( obs[i][j+1] - obs[i][j-1] ) / dx;
            psi_y = 0.5*( obs[i+1][j] - obs[i-1][j] ) / dx;
            r1 = (double) (x00 - j);
            r2 = (double) (y00 - i);
            rs = sqrt(r1*r1 + r2*r2);
            if (rs > 0)
                h = (r1*psi_x + r2*psi_y) / rs;
            else
                h = (r1*psi_x + r2*psi_y);
            
            v[i][j] = max(obs[i][j], h);
        }
    
    //... and at the boundaries ...
    // Neumann b.c.
    for(i=0; i<N; i++) {
        v[i][0] = v[i][1];
        v[i][N-1] = v[i][N-2];
        
        v[0][i] = v[1][i];
        v[N-1][i] = v[N-2][i];
    }
    
    /*
    for(i=1; i<N-1; i++) {
        // case j=0
        psi_x =     ( obs[i][1]   - obs[i][0] )   / dx;
        psi_y = 0.5*( obs[i+1][0] - obs[i-1][0] ) / dx;
        r1 = (double) (x00 - 0);
        r2 = (double) (y00 - i);
        rs = sqrt(r1*r1 + r2*r2);
        h = (r1*psi_x + r2*psi_y) / rs;
        v[i][0] = max(obs[i][0], h);
        // case j=N-1
        psi_x =     ( obs[i][N-1]   - obs[i][N-2] )   / dx;
        psi_y = 0.5*( obs[i+1][N-1] - obs[i-1][N-1] ) / dx;
        r1 = (double) (x00 - N+1);
        r2 = (double) (y00 - i);
        rs = sqrt(r1*r1 + r2*r2);
        h = (r1*psi_x + r2*psi_y) / rs;
        v[i][N-1] = max(obs[i][0], h);
        // case i=0
        psi_y =     ( obs[1][i]   - obs[0][i] )   / dx;
        psi_x = 0.5*( obs[0][i+1] - obs[0][i-1] ) / dx;
        r1 = (double) (x00 - i);
        r2 = (double) (y00 - 0);
        rs = sqrt(r1*r1 + r2*r2);
        h = (r1*psi_x + r2*psi_y) / rs;
        v[0][i] = max(obs[i][0], h);
        // case i=N-1
        psi_y =     ( obs[N-1][i]   - obs[N-2][i] )   / dx;
        psi_x = 0.5*( obs[N-1][i+1] - obs[N-1][i-1] ) / dx;
        r1 = (double) (x00 - i);
        r2 = (double) (y00 - N+1);
        rs = sqrt(r1*r1 + r2*r2);
        h = (r1*psi_x + r2*psi_y) / rs;
        v[N-1][i] = max(obs[i][0], h);
    }
    
    //... and at the corners.
    psi_x =     ( obs[0][1] - obs[0][0] ) / dx;
    psi_y =     ( obs[1][0] - obs[0][0] ) / dx;
    r1 = (double) (x00 - 0);
    r2 = (double) (y00 - 0);
    rs = sqrt(r1*r1 + r2*r2);
    h = (r1*psi_x + r2*psi_y) / rs;
    v[0][0] = max(obs[0][0], h);
    
    psi_x =     ( obs[0][N-1] - obs[0][N-2] ) / dx;
    psi_y =     ( obs[1][N-1] - obs[0][N-1] ) / dx;
    r1 = (double) (x00 - N+1);
    r2 = (double) (y00 - 0);
    rs = sqrt(r1*r1 + r2*r2);
    h = (r1*psi_x + r2*psi_y) / rs;
    v[0][N-1] = max(obs[0][N-1], h);
    
    psi_y =     ( obs[N-1][0] - obs[N-2][0] ) / dx;
    psi_x =     ( obs[N-1][1] - obs[N-1][0] ) / dx;
    r1 = (double) (x00 - 0);
    r2 = (double) (y00 - N+1);
    rs = sqrt(r1*r1 + r2*r2);
    h = (r1*psi_x + r2*psi_y) / rs;
    v[N-1][0] = max(obs[N-1][0], h);
    
    psi_y =     ( obs[N-1][N-1] - obs[N-2][N-1] ) / dx;
    psi_x =     ( obs[N-1][N-1] - obs[N-1][N-2] ) / dx;
    r1 = (double) (x00 - N+1);
    r2 = (double) (y00 - N+1);
    rs = sqrt(r1*r1 + r2*r2);
    h = (r1*psi_x + r2*psi_y) / rs;
    v[N-1][N-1] = max(obs[N-1][N-1], h);
    */
    
    
    // compute visibility function of the auxiliary function  
    singleVisFunc(visShadTemp, y00, x00, v, N, dx);
    
    // FOR TESTING:
//     for (i=0;i<N;i++)
//         for(j=0;j<N;j++) {
//             visShadTemp[i][j] = v[i][j];
//         }
            
    
    
    // delete v;
	for(i=0; i < N; i++)
		free(v[i]);
	free(v);
    
}


void singleVisFunc(double** visFuncTemp,int x00, int y00,double** obs,int N,double dx) {
    if (x00 >= 0 && x00 < N && y00 >= 0 && y00 < N) {
        // boundary condition
        visFuncTemp[x00][y00] = obs[x00][y00];

        // sweep in each quadrant
        singleSweep(visFuncTemp,x00,y00,obs,N,dx, 1, 1);
        singleSweep(visFuncTemp,x00,y00,obs,N,dx,-1, 1);
        singleSweep(visFuncTemp,x00,y00,obs,N,dx, 1,-1);
        singleSweep(visFuncTemp,x00,y00,obs,N,dx,-1,-1);
    }              
}



void singleSweep(double** visFuncTemp,int x00,int y00,double** obs,int N,double dx,int sX, int sY) 
{
    
    int i,j, ssX,ssY;
    double r1,r2,rs, De,Nu, v_temp;
    
    
    for( i=x00; (sX<0 ? i>=0:i<N); i+=sX )
		for( j=y00; (sY<0 ? j>=0:j<N); j+=sY ) {
            
            if ((i==x00)&& (j==y00)) {
                // do nothing
            } else {
                r1 = (double) (i-x00);
                r2 = (double) (j-y00);
            
                //rs = sqrt(r1*r1 + r2*r2)*dx;
            
                r1 = (double) sX*r1;///rs;
                r2 = (double) sY*r2;///rs;
               
                ssX = i-sX;
                if ( (ssX < 0) || (ssX > N-1) )
                    ssX = i;
                ssY = j-sY;
                if ( (ssY < 0) || (ssY > N-1) )
                    ssY = j;
                
                De = r1 + r2;    
                Nu = r1*visFuncTemp[ssX][j] + r2*visFuncTemp[i][ssY];
                
                v_temp = Nu / De;
                visFuncTemp[i][j] = min(v_temp, obs[i][j]);
            }
        }
}    
