   /*
  * =====================================================================================
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation, either version 3 of the License, or
  *  (at your option) any later version.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *
  * -------------------------------------------------------------------------------------
  *  The primary purpose in distributing this source code is to enable 
  *  readers to reproduce the numerical results reported in our manuscript
  *  on optimal control under constraints on reset-renewable resources.
  *  See  http://arxiv.org/abs/1110.6221
  *
  *  Please see Readme.txt for instructions on configuring/running this program.
  * -------------------------------------------------------------------------------------
  *  Problem statement by R. Takei.
  *  The idea of the algorithm by A. Vladimirsky.
  *  The numerical method development by R. Takei & A. Vladimirsky.
  *  Implementation by R. Takei.
  *  
  *  The method implemented in these files is not as general as the description 
  *  in the paper and relies on several implicit assumptions:
  *
  *		1) the dynamics is isotropic and the primary cost is identically = 1 on Omega;
  *		2) the secondary cost on U is identically = 1;
  *		3) the discretization parameters (h and deltaB) satisfy
  *			a CFL-type condition:  deltaB * f(x) \leq h.
  *
  *  These assumptions are not essential for the approach, but simplify the implementation.
  *
  *  To handle inhomogenious/anisotropic secondary costs and arbitrary deltaB/h ratios,
  *  an alternative implementation was developed & tested by W. Chen, Z. Clawson, 
  *  and S. Kirov based on the code for the U=Omega case previously developed by A. Kumar.
  *  This more general implementation is not included in the current distribution.
  * =====================================================================================
  */

#include <math.h>
#include "matrix.h"
#include "mex.h"   //--This one is required
#include "PositiveDistanceSolver.h"
#include "VisSweep.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    //---Inside mexFunction---
    //Declarations
    double *obsobs,*outArray1;
    double **obs,**u,**v;
    const int *dims;
    double xE,yE;
    int x00,y00, xE_0,yE_0;
	int N, i,j, NumIter;
    double L,hx,err; // assume equi-spaced in x and y
    
    
    
    //Get the input
    obsobs = mxGetPr(prhs[0]);
    xE     = mxGetScalar(prhs[1]); 
    yE     = mxGetScalar(prhs[2]);
    dims   = mxGetDimensions(prhs[0]);
    N      = dims[0];
    
    
    // domain is [-L,L]^2.
    L = 1;
    hx = 2*L / (N-1); 
    // E's pixel location
    xE_0 = (int) xE - 1;
    yE_0 = (int) yE - 1;
    
    // memory allocation for u and Obstacles
    u   = (double **) malloc ( N * sizeof(double*));
    obs = (double **) malloc ( N * sizeof(double*));
	for (i=0;i<N;i++) {
        u[i]   = (double *) malloc (N * sizeof(double));
        obs[i] = (double *) malloc (N * sizeof(double));
    }
    
    // assignment u and obstacle values 
    for (i=0; i < N; i++) {
		for (j=0; j < N; j++) {
            u[i][j] = 0;
            obs[i][j] = obsobs[(j*N)+i];
		}
	}
    
    // compute visibility function from E's initial point, store in v
    singleVisFunc(u, yE_0, xE_0, obs, N, hx);
    /*
    // solve signed distance PDE (for short time)
    NumIter = 50;
    signDistPDE(u, N, L, NumIter);

    for(i=0; i<N; i++)
        for(j=0; j<N; j++) {
            if (u[i][j] >= hx)
                u[i][j] = 100;
            if (u[i][j] < -hx)
                u[i][j] = -hx;
        }

    // sweep only in the positive part
    for(i=0; i<100; i++) {
        err = sweepSignDist(u, obs, N);
        if (err < hx*0.5) 
            break;
    }

    for(i=0; i<5; i++) {
        err = FSM_HO(u, obs, N);
        //printf("error = %f\n",err);
        if (err < hx*0.1){
            break;
        }
    }
    */
    //Allocate memory and assign output pointer
    plhs[0] = mxCreateDoubleMatrix(N, N, mxREAL); //mxReal is our data-type
    //Get a pointer to the data space in our newly allocated memory
    outArray1 = mxGetPr(plhs[0]);
    
    for (i=0; i < N; i++) {
		for (j=0; j < N; j++) {
            outArray1[(j*N)+i] = u[i][j];
		}
	}
    
    // delete u;
	for(i=0; i < N; i++) {
		free(u[i]);
        free(obs[i]);
	}
	free(u);
    free(obs);
    
}
