   /*
  * =====================================================================================
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation, either version 3 of the License, or
  *  (at your option) any later version.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *
  * -------------------------------------------------------------------------------------
  *  The primary purpose in distributing this source code is to enable 
  *  readers to reproduce the numerical results reported in our manuscript
  *  on optimal control under constraints on reset-renewable resources.
  *  See  http://arxiv.org/abs/1110.6221
  *
  *  Please see Readme.txt for instructions on configuring/running this program.
  * -------------------------------------------------------------------------------------
  *  Problem statement by R. Takei.
  *  The idea of the algorithm by A. Vladimirsky.
  *  The numerical method development by R. Takei & A. Vladimirsky.
  *  Implementation by R. Takei.
  *  
  *  The method implemented in these files is not as general as the description 
  *  in the paper and relies on several implicit assumptions:
  *
  *		1) the dynamics is isotropic and the primary cost is identically = 1 on Omega;
  *		2) the secondary cost on U is identically = 1;
  *		3) the discretization parameters (h and deltaB) satisfy
  *			a CFL-type condition:  deltaB * f(x) \leq h.
  *
  *  These assumptions are not essential for the approach, but simplify the implementation.
  *
  *  To handle inhomogenious/anisotropic secondary costs and arbitrary deltaB/h ratios,
  *  an alternative implementation was developed & tested by W. Chen, Z. Clawson, 
  *  and S. Kirov based on the code for the U=Omega case previously developed by A. Kumar.
  *  This more general implementation is not included in the current distribution.
  * =====================================================================================
  */

#include <math.h>
#include "matrix.h"
#include "mex.h"   //--This one is required
#include "sweepProc.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    
    //Declarations
    double *uValues, *bdryCondValues, *obsValues, *speedValues;
    double **u, **bdryCond, **obs, **speed;
    const int *dims;
	int N, iterations;
    int i,j;
    double h;
    
    double maxError;
    double TOL = 0.000001;
    
    //Get the input
    uValues = mxGetPr(prhs[0]);
    dims    = mxGetDimensions(prhs[0]);
    N       = (int) dims[0];
    
    bdryCondValues = mxGetPr(prhs[1]);
    obsValues      = mxGetPr(prhs[2]);
    iterations     = (int) mxGetScalar(prhs[3]);
    h              =       mxGetScalar(prhs[4]);
    speedValues    = mxGetPr(prhs[5]);
    
    // memory allocation for u and Contaminated
    bdryCond = (double **) malloc ( N * sizeof(double*));
    obs      = (double **) malloc ( N * sizeof(double*));
    u        = (double **) malloc ( N * sizeof(double*));
    speed    = (double **) malloc ( N * sizeof(double*));
    
	for (i=0;i<N;i++) {
        bdryCond[i] = (double *) malloc (N * sizeof(double));
        obs[i]      = (double *) malloc (N * sizeof(double));
        u[i]        = (double *) malloc (N * sizeof(double));
        speed[i]    = (double *) malloc (N * sizeof(double));
	}
    
    // assignment u and Contaminated values 
    for (i=0; i < N; i++)
		for (j=0; j < N; j++) {
            u[i][j]        = uValues[(j*N)+i];
            obs[i][j]      = obsValues[(j*N)+i];
            bdryCond[i][j] = bdryCondValues[(j*N)+i];
            speed[i][j]    = speedValues[(j*N)+i];
        }
    
    // run the algorithm
    
    for (i=0;i<iterations;i++) {
        maxError =  eikonalSweep(u, obs, bdryCond, N, h, speed);
        if (maxError < TOL) {
            //printf("Other Eikonal solver: break at iteration %d\n",i);
            break;
        }
    }
    // send the processed u to the output
    for (i=0; i < N; i++)
		for (j=0; j < N; j++)
                uValues[(j*N)+i] = u[i][j];
    

       
    // Output
    //plhs[0] = (double *) prhs[0]; 
    
    //delete data
	for(i=0; i< N; i++)
    {
		free(u[i]);
        free(bdryCond[i]);
        free(obs[i]);
        free(speed[i]);
    }
    free(u);
    free(bdryCond);
    free(obs);
    free(speed);
    
}
