   /*
  * =====================================================================================
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation, either version 3 of the License, or
  *  (at your option) any later version.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *
  * -------------------------------------------------------------------------------------
  *  The primary purpose in distributing this source code is to enable 
  *  readers to reproduce the numerical results reported in our manuscript
  *  on optimal control under constraints on reset-renewable resources.
  *  See  http://arxiv.org/abs/1110.6221
  *
  *  Please see Readme.txt for instructions on configuring/running this program.
  * -------------------------------------------------------------------------------------
  *  Problem statement by R. Takei.
  *  The idea of the algorithm by A. Vladimirsky.
  *  The numerical method development by R. Takei & A. Vladimirsky.
  *  Implementation by R. Takei.
  *  
  *  The method implemented in these files is not as general as the description 
  *  in the paper and relies on several implicit assumptions:
  *
  *		1) the dynamics is isotropic and the primary cost is identically = 1 on Omega;
  *		2) the secondary cost on U is identically = 1;
  *		3) the discretization parameters (h and deltaB) satisfy
  *			a CFL-type condition:  deltaB * f(x) \leq h.
  *
  *  These assumptions are not essential for the approach, but simplify the implementation.
  *
  *  To handle inhomogenious/anisotropic secondary costs and arbitrary deltaB/h ratios,
  *  an alternative implementation was developed & tested by W. Chen, Z. Clawson, 
  *  and S. Kirov based on the code for the U=Omega case previously developed by A. Kumar.
  *  This more general implementation is not included in the current distribution.
  * =====================================================================================
  */

#define min(a,b) (((a) < (b)) ? (a) : (b))
#define max(a,b) (((a) > (b)) ? (a) : (b))
#define abs(a) (((a) < 0) ? (-a) : (a))
#define PI 3.14159265358979323846
#define SQRT2 1.414213562373
#define SQRT5 2.236067977499

double OtherOptEikonalSweep(double** u, double** udist, double** ubdry, double** obs, int N, double h, double tau, double** speed);
void ContaminatedEikonalSweep(double ***u, double*** URbdry, double** obs, double*** bdryCond, int N, int Nb, double h, double hb, double** speed, double** unV);
double eikonalSweep(double **u, double** obs, double** bdryCond, int N, double h, double** speed);
double ZeroEikonalSweep(double **u, double** obs, double** bdryCond, int N, double h);

double goldenSectionSearch(double a,double c,double b, double prec, double w1,double w2,double w3,double w4, double r);
double Fk(double t, double r, double w1, double w2, double w3, double w4);

double OtherOptEikonalSweep(double** u, double** udist, double** ubdry, double** obs, int N, double h, double tau, double** speed) {
   
    int s1,s2, i,j;
    double norm, ux,uy, dx,dy, u_new, u0, ax,ay, udx,udy, c;
    
    double maxError = 0;
    double currError;
    
    for (s1=1; s1>=-1; s1-=2)
    for (s2=1; s2>=-1; s2-=2)
    for( i=(s1<0 ? N -1 : 0 ); (s1<0 ? i >= 0 : i <= N -1); i+=s1 )
    for( j=(s2<0 ? N -1 : 0 ); (s2<0 ? j >= 0 : j <= N -1); j+=s2 )
    {
        if ( obs[i][j] > 0.5 ) {
            // don't do anything in obstacle
        } else if ( ubdry[i][j] >= 0 ) {
            //set boundary conditions
            u[i][j] = ubdry[i][j];
        } else {
            
            c = 1/speed[i][j];
            
            u0 = udist[i][j];
            
            
            if(j==0) {   
                ux = u[i][j+1]; udx = udist[i][j+1]; 
            } else if (j==N-1) {
                ux = u[i][j-1]; udx = udist[i][j-1];
            } else {
                if (udist[i][j+1] < udist[i][j-1]) {
                    ux = u[i][j+1]; udx = udist[i][j+1]; 
                } else {
                    ux = u[i][j-1]; udx = udist[i][j-1]; 
                } 
            }
            
            if(i==0) {
                uy = u[i+1][j]; udy = udist[i+1][j]; 
            } else if (i==N-1) {
                uy = u[i-1][j]; udy = udist[i-1][j];
            } else {
                if (udist[i+1][j] < udist[i-1][j]) {
                    uy = u[i+1][j]; udy = udist[i+1][j]; 
                } else {
                    uy = u[i-1][j]; udy = udist[i-1][j]; 
                } 
            }
            
            dx = udx - u0;
            dy = udy - u0;
            
            norm = sqrt(dx*dx + dy*dy);
            if ( norm == 0) { 
                ax = tau/SQRT2;    ay = tau/SQRT2;
            } else {
                ax = tau*dx/norm;  ay = tau*dy/norm;
            }
            
            // update formula
            u_new = (h*tau*c + ux*abs(ax) + uy*abs(ay)) / (abs(ax) + abs(ay));
            // record error
            currError = u[i][j] - u_new;
            currError = abs(currError);
            maxError = max(maxError, currError);
            
            u[i][j] = min( u[i][j] , u_new);
        } 
    }
    // return max error
    return maxError;
    
}

void ContaminatedEikonalSweep(double ***u, double*** URbdry, double** obs, double*** bdryCond, int N, int Nb, double h, double hb, double** speed, double** unV) {
    int i,j,k, II, JJ;
    double u_new, w1,w2,w3,w4;
    
    double tau = hb;
    double r   = tau/h;
    double prec, ts, minFk;
    
    double u_temp, ux,uy, gradU, u0, c;
    
    for (i=0; i<N; i++)
    for (j=0; j<N; j++) { 
        
        if ( obs[i][j] > 0.5 ) {
            // do nothing in obstacles
        } else if (bdryCond[i][j][0] >= 0) {
            u[i][j][0] = bdryCond[i][j][0]; //set u to boundary condition
        } else if (URbdry[i][j] >= 0) {
            u[i][j][0] = URbdry[i][j][0]; // set to UR boundary values
        }
    }
    
    // sweep 'up'
    for (k=1; k < Nb; k++) {
        for (i=0; i<N; i++)
        for (j=0; j<N; j++) {    
            if ( obs[i][j] > 0.5 ) {
                // do nothing in obstacles
            } else if (bdryCond[i][j][k] >= 0) {
                u[i][j][k] = bdryCond[i][j][k]; //set u to boundary condition
            } else if (URbdry[i][j][k] >= 0) {
                u[i][j][k] = URbdry[i][j][k]; // set to UR boundary values
            } else if (u[i][j][k-1] <= unV[i][j]) {
                u[i][j][k] = unV[i][j];
            } else {
                
                c = 1/speed[i][j];
                
                if(j==0) {   
                	ux = u[i][j+1][k-1]; JJ = 1; 
                } else if (j==N-1) {
                    ux = u[i][j-1][k-1]; JJ = -1; 
                } else {
                    if (u[i][j+1][k-1] < u[i][j-1][k-1]) {
                        ux = u[i][j+1][k-1]; JJ = 1; 
                    } else {
                        ux = u[i][j-1][k-1]; JJ = -1;    }
                }
                
                if(i==0) {
                    uy = u[i+1][j][k-1]; II = 1; 
                } else if (i==N-1) {
                    uy = u[i-1][j][k-1]; II = -1;
                } else {
                    if (u[i+1][j][k-1] < u[i-1][j][k-1]) { 
                        uy = u[i+1][j][k-1]; II = 1; 
                    } else {
                        uy = u[i-1][j][k-1]; II = -1;    }
                }
                
				{
                    //SEMI-LAGRANGIAN SCHEME:
                    /* Set up simplex as follows:
                     *
                     *    w2 ---- w3=u[i+II][j+JJ][k-1]  
                     *    |       |
                     *    |       |
                     *   w1=u0 -- w4
                     *
                     * and find minimum on a quarter circle of radius r=tau/h 
                     * centered at the w1 location. 
                     * The values of w2 and w4 will depend on JJ, II.
                     */
                    w1 = u[i   ][j   ][k-1];
                    w3 = u[i+II][j+JJ][k-1];    
                    if (II*JJ == 1) {
                        w2 = u[i+II][j   ][k-1];  
                        w4 = u[i   ][j+JJ][k-1];
                    } else {
                        w2 = u[i   ][j+JJ][k-1];  
                        w4 = u[i+II][j   ][k-1];
                    }

                    prec = ((PI*0.5)/90)/20; // precision
                    // do golden section search
                    ts = goldenSectionSearch(0,PI*(3-sqrt(5.))*0.25,PI*0.5, prec, w1,w2,w3,w4, r);
                    // extract minimum value
                    minFk = Fk(ts,r, w1,w2,w3,w4);
                    
                    
                    // update semi-Lagrangian step
                    u_new =  tau*c + minFk;
                    u[i][j][k] = min(u_new, u[i][j][k]);
                }
            } 
        }
    }
}

// recursive implementation of the golden section search for finding the minimizer
double goldenSectionSearch(double a,double c,double b, double prec, double w1,double w2,double w3,double w4, double r) {
    
    double diff = a-b;
    // check if precision is met
    if (abs(diff) < prec) 
        return (a+b)*0.5;
    // new point
    double d = c + (3-sqrt(5.))*0.5*(b-c);
    
    // choose domain for recursion
    if (Fk(d,r,w1,w2,w3,w4) < Fk(c,r,w1,w2,w3,w4)) 
        return goldenSectionSearch(c,d,b,prec, w1,w2,w3,w4, r);
    else 
        return goldenSectionSearch(d,c,a,prec, w1,w2,w3,w4, r);
}

// bilinear interpolation function in simplex
double Fk(double t, double r, double w1, double w2, double w3, double w4) {
    double g1 = r*cos(t);
    double g2 = r*sin(t);
    
    return (1-g2)*(1-g1)*w1 + g2*(1-g1)*w2 + g2*g1*w3 + (1-g2)*g1*w4;
}

double eikonalSweep(double **u, double** obs, double** bdryCond, int N, double h, double** speed) {
    
    int s1,s2, i,j;
    double DET, ux,uy, u_new, c;
    
    double maxError = 0;
    double currError;
    
    for (s1=1; s1>=-1; s1-=2)
    for (s2=1; s2>=-1; s2-=2)
    for( i=(s1<0 ? N -1 : 0 ); (s1<0 ? i >= 0 : i <= N -1); i+=s1 )
    for( j=(s2<0 ? N -1 : 0 ); (s2<0 ? j >= 0 : j <= N -1); j+=s2 )
    {
        if ( obs[i][j] > 0.5 ) {
            // don't do anything in obstacle
        } else if ( bdryCond[i][j] >= 0 ) {
            // set boundary condition
            u[i][j] = bdryCond[i][j];
        } else if ( bdryCond[i][j] <  0 ) {
            
            c = 1/speed[i][j];
            
            if(i==0) {
                uy = u[i+1][j];
            } else if (i==N-1) {
                uy = u[i-1][j];
            } else {
                uy = min(u[i+1][j], u[i-1][j]);
            }
            
            if(j==0) {
                ux = u[i][j+1];
            } else if (j==N-1) {
                ux = u[i][j-1];
            } else {
                ux = min(u[i][j+1], u[i][j-1]);
            }
            
            
            // general update scheme
            DET = 2*h*h*c*c - (ux-uy)*(ux-uy);
            if (DET < 0)
                u_new = h*c + min(ux,uy);
            else  {  
                u_new = 0.5*(ux + uy + sqrt(DET));
                
                if (uy > u_new )
                   u_new = h*c + ux; 
                else if (ux > u_new)
                   u_new = h*c + uy;
            }
            
            // record error
            currError = u[i][j] - u_new;
            currError = abs(currError);
            maxError  = max(maxError, currError);
            
            // check with current u
            u[i][j] = min(u[i][j], u_new);
        }
    }
    //return error
    return maxError;
}

double ZeroEikonalSweep(double **u, double** obs, double** bdryCond, int N, double h) {
    
    int s1,s2, i,j;
    double DET, ux,uy, u_new;
    
    double maxError = 0;
    double currError;
    
    for (s1=1; s1>=-1; s1-=2)
    for (s2=1; s2>=-1; s2-=2)
    for( i=(s1<0 ? N -1 : 0 ); (s1<0 ? i >= 0 : i <= N -1); i+=s1 )
    for( j=(s2<0 ? N -1 : 0 ); (s2<0 ? j >= 0 : j <= N -1); j+=s2 )
    {
        if ( obs[i][j] > 0.5 ) {
            // don't do anything in obstacle
        } else if ( bdryCond[i][j] >= 0 ) {
            // set boundary condition
            u[i][j] = bdryCond[i][j];
        } else if ( bdryCond[i][j] <  0 ) {
            
            if(i==0) {
                uy = u[i+1][j];
            } else if (i==N-1) {
                uy = u[i-1][j];
            } else {
                uy = min(u[i+1][j], u[i-1][j]);
            }
            
            if(j==0) {
                ux = u[i][j+1];
            } else if (j==N-1) {
                ux = u[i][j-1];
            } else {
                ux = min(u[i][j+1], u[i][j-1]);
            }
            
            u_new = min(ux,uy);
            
            // record error
            currError = u[i][j] - u_new;
            currError = abs(currError);
            maxError  = max(maxError, currError);
            
            // check with current u
            u[i][j] = min(u[i][j], u_new);
        }
    }
    //return error
    return maxError;
}

