% =====================================================================================
%
%  This program is free software: you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation, either version 3 of the License, or
%  (at your option) any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
% -------------------------------------------------------------------------------------
%  The primary purpose in distributing this source code is to enable 
%  readers to reproduce the numerical results reported in our manuscript
%  on optimal control under constraints on reset-renewable resources.
%  See  http://arxiv.org/abs/1110.6221
%
%  Please see Readme.txt for instructions on configuring/running this program.
% -------------------------------------------------------------------------------------
%  Problem statement by R. Takei.
%  The idea of the algorithm by A. Vladimirsky.
%  The numerical method development by R. Takei & A. Vladimirsky.
%  Implementation by R. Takei.
%  
%  The method implemented in these files is not as general as the description 
%  in the paper and relies on several implicit assumptions:
%
%		1) the dynamics is isotropic and the primary cost is identically = 1 on Omega;
%		2) the secondary cost on U is identically = 1;
%		3) the discretization parameters (h and deltaB) satisfy
%			a CFL-type condition:  deltaB * f(x) \leq h.
%
%  These assumptions are not essential for the approach, but simplify the implementation.
%
%  To handle inhomogenious/anisotropic secondary costs and arbitrary deltaB/h ratios,
%  an alternative implementation was developed & tested by W. Chen, Z. Clawson, 
%  and S. Kirov based on the code for the U=Omega case previously developed by A. Kumar.
%  This more general implementation is not included in the current distribution.
% =====================================================================================
% This work was supported by
% ONR grants N0014-03-1-0071, N00014-07-1-0810, N00014-08-1-1119,
% DOE grant DE-FG02-05ER25710,
% ARO MURI through Rice University and the CHASE MURI grant 556016,
% National Science Foundation grants DMS-0514487 and DMS-1016150,
% and the NSF sponsored REU'2010 at Cornell University.
% =====================================================================================

% THIS FUNCTION IS TO BE CALLED BY 'ContaminatedEikonal.m'!


function [R,OB,f,xT,yT,x,y] = testcases(number,L,N)

% generate mesh grid
[x,y] = meshgrid(linspace(-L,L,N));


if (number == 1) % test = 1: generic example with target in safe set
    xT = -0.7;
    yT = -0.4;
    
    R = max(0.2*abs(x),abs(y+0.3+0.2)) - 0.1;
    R = min(R, max(abs(x-0.4),0.3*abs(y+0.2)) - 0.1);
    R = min(R, max(0.2*abs(x),abs(y-0.3+0.2)) - 0.1);
    R = min(R, max(abs(x+0.4),0.9*abs(y-0.15+0.2)) - 0.1);
    R = min(R, max(abs(x),abs(y-0.5))-0.15);      % "island" 1
    R = min(R, max(abs(x-0.5),abs(y-0.7))-0.15); % "island" 2
    
    obs_ls = ones(size(x)); % no obstacle
    f = ones(N,N); % unit speed
elseif (number == 2) % test = 2: high cost in safe set
    xT = -0.7;
    yT = -0.4;
    
    R = max(0.2*abs(x),abs(y+0.3+0.2)) - 0.1;
    R = min(R, max(abs(x-0.4),0.3*abs(y+0.2)) - 0.1);
    R = min(R, max(0.2*abs(x),abs(y-0.3+0.2)) - 0.1);
    R = min(R, max(abs(x+0.4),0.9*abs(y-0.15+0.2)) - 0.1);
    R = min(R, max(abs(x),abs(y-0.5))-0.15);      % "island" 1
    R = min(R, max(abs(x-0.5),abs(y-0.7))-0.15); % "island" 2
    
    obs_ls = ones(size(x)); % no obstacle
    f = double(R<0).*0.3 + double(R>=0).*1.0;
elseif (number == 3) % test = 3: 2-d sinusoidal cost function
    xT = -0.7;
    yT = -0.4;
    
    R = max(0.2*abs(x),abs(y+0.3+0.2)) - 0.1;
    R = min(R, max(abs(x-0.4),0.3*abs(y+0.2)) - 0.1);
    R = min(R, max(0.2*abs(x),abs(y-0.3+0.2)) - 0.1);
    R = min(R, max(abs(x+0.4),0.9*abs(y-0.15+0.2)) - 0.1);
    R = min(R, max(abs(x),abs(y-0.5))-0.15);      % "island" 1
    R = min(R, max(abs(x-0.5),abs(y-0.7))-0.15); % "island" 2
    
    obs_ls = ones(size(x)); % no obstacle
    f = 0.5*(2 - sin(5*pi*x).*sin(5*pi*y));
elseif (number == 4) % test = 4: square chain of square
    xT = -0.5;
    yT = -0.5;
    
    F = zeros(N,N,8);
    F(:,:,1) = max(abs(x+0.5),abs(y+0.5)) - 0.2;
    F(:,:,2) = max(abs(x+0.0),abs(y+0.5)) - 0.2;
    F(:,:,3) = max(abs(x-0.5),abs(y+0.5)) - 0.2;
    F(:,:,4) = max(abs(x-0.5),abs(y+0.0)) - 0.2;
    F(:,:,5) = max(abs(x-0.5),abs(y-0.5)) - 0.2;
    F(:,:,6) = max(abs(x+0.0),abs(y-0.5)) - 0.2;
    F(:,:,7) = max(abs(x+0.5),abs(y-0.5)) - 0.2;
    F(:,:,8) = max(abs(x+0.5),abs(y+0.0)) - 0.2;
    
    R = min(F,[],3);
    
    obs_ls = ones(size(x)); % no obstacle
    
    OS = max(abs(x),abs(y)) - 0.7;
    IS = max(abs(x),abs(y)) - 0.3;
    SS = max(abs(x+0.5),4*abs(y+0.25)) - 0.2;
    LOW = double(max(OS, max(-IS, -SS)) > 0);
    HIGH = double(LOW == 0);
    f = 10*HIGH + 0.1*LOW; 
elseif (number == 5) % test = 5: for convergence test
    xT = 1.0;
    yT = 0.0;
    
    R = (x - 1/3);  
    
    obs_ls = ones(size(x)); % no obstacle
    f = ones(N,N); % unit speed
elseif (number == 6) % test = 6: visibility example
    xT = -0.5;
    yT =  0.9;
    
    obs_ls = sqrt(x.^2+y.^2) - 0.3;
    obs_ls = min(obs_ls, abs(x+0.4)+abs(y-0.6)- 0.2);
    obs_ls = min(obs_ls, max(1.4*abs(x-0.5),abs(y+0.4))- 0.2);
    obs_ls = min(obs_ls, sqrt((x+0.7).^2+(y-0.15).^2)- 0.05);
    
    jP = round(N*0.9);
    iP = round(N*0.9);
    R = computeVisibility(obs_ls, jP,iP);
    
    f = ones(N,N); % unit speed
end

OB = double( obs_ls < 0 );